#include <stdafx.h>
#include <comutil.h>
#include <comadmin.h>
#include <assert.h>

#include "macro.h"
#include "Adminwrap.h"
#include <atlbase.h>
//*********************************************
// Globals

WCHAR g_wczAPP[]          = L"Applications";
WCHAR g_wczCOMP[]         = L"Components";
WCHAR g_wczNAME[]         = L"Name";
WCHAR g_wczProgIDProp[]   = L"ProgID";
WCHAR g_wczSUB[]          = L"SubscriptionsForComponent";
WCHAR g_wczTRANSUB[]      = L"TransientSubscriptions";
WCHAR g_wczPUBPROP[]      = L"PublisherProperties";
WCHAR g_wczTRANSPUBPROP[] = L"TransientPublisherProperties";

_bstr_t g_bAPP            = g_wczAPP;
_bstr_t g_bCOMP           = g_wczCOMP;
_bstr_t g_bPUBID          = L"PublisherID";
_bstr_t g_bFILTER         = L"MultiInterfacePublisherFilterCLSID";
_bstr_t g_bALLOWINPROC    = L"AllowInprocSubscribers";
_bstr_t g_bFIREINPARALLEL = L"FireInParallel";
_bstr_t g_bSUB            = g_wczSUB;
_bstr_t g_bNAME           = L"Name";
_bstr_t g_bVALUE          = L"Value";
_bstr_t g_bECID           = L"EventCLSID";
_bstr_t g_bIID            = L"InterfaceID";
_bstr_t g_bMETHOD         = L"MethodName";
_bstr_t g_bCRITERIA       = L"FilterCriteria";
_bstr_t g_bPUBPROP        = g_wczPUBPROP;
_bstr_t g_bTRANSSUB       = g_wczTRANSUB;
_bstr_t g_bSUBINT         = L"SubscriberInterface";
_bstr_t g_bTRANSPUBPROP   = g_wczTRANSPUBPROP;


//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::CApp
// Purppose    : CApp constructor
// Parameters  :
//             : ICOMAdminCatalog* pcat - Pointer to ICOMAdminCatalog interface
//             : WCHAR* pwzAppName         - Name of the Application
///////////////////////////////////////////////////////////////////////////////////////
CApp::CApp()
{
   HRESULT hr = ::CoCreateInstance(__uuidof(COMAdminCatalog), NULL, CLSCTX_SERVER, __uuidof(ICOMAdminCatalog),(LPVOID*) &m_pAdminCatalog);
   ASSERT(SUCCEEDED(hr));
}
CApp::CApp(ICOMAdminCatalog* pAdminCatalog, WCHAR* pwzAppName)
{
   if (pAdminCatalog)
   {
      m_pAdminCatalog = pAdminCatalog;
      m_pAdminCatalog->AddRef();
   }

   m_pwzAppName = new WCHAR[wcslen(pwzAppName)+1];

   wcscpy(m_pwzAppName, pwzAppName); 
}

//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::~CApp
// Purppose    : CApp destructor
///////////////////////////////////////////////////////////////////////////////////////
CApp::~CApp()
{
   SAFERELEASE(m_pAdminCatalog);
   if (m_pwzAppName)
   {
      DeleteApp();
      delete [] m_pwzAppName;
      m_pwzAppName=0;
   }
}

//////////////////////////////////////////////////////////////////////////////////////
// Name         : CApp::SetProperties
// Purppose     : Sets the properties of the CApp object
// Parameters   :
//              : ICOMAdminCatalog* pcat - Pointer to ICOMAdminCatalog interface
//              : WCHAR* pwzAppName         - Name of the Application
// Return Value : HRESULT
///////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::SetProperties(ICOMAdminCatalog* pAdminCatalog, WCHAR* pwzAppName)
{
   if (!pAdminCatalog || !pwzAppName)
      return E_POINTER;

   SAFERELEASE(m_pAdminCatalog);
   m_pAdminCatalog = pAdminCatalog;
   m_pAdminCatalog->AddRef();

   if (m_pwzAppName)
   {
      DeleteApp();
      delete [] m_pwzAppName;
      m_pwzAppName=0;
   }

   m_pwzAppName = new WCHAR[wcslen(pwzAppName)+1];

   wcscpy(m_pwzAppName, pwzAppName);
   return S_OK;
}

//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::CreateApp
// Purppose    : Creates a COM+ Application
// Parameters  :
//             : BOOL bServer - Specifies if is a Server app of Library app
//             : BOOL bQueued - Specifies if the app is QUEUED or not
// Return Value : HRESULT
///////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::CreateApp(BOOL bServer, BOOL bQueued)
{
   HRESULT           hr =0;
   ICatalogObject    *pobj = 0;
   ICatalogCollection   *pcoll = 0;
   IDispatch         *pdisp = 0;
   _bstr_t           bname; ;
   long           lret =0;
   VARIANT           var;

   if (!m_pwzAppName || !m_pAdminCatalog)
      return E_POINTER;

   hr = GetCollection(m_pAdminCatalog, g_bAPP, &pcoll);
   exit(hr);

   hr = pcoll->Add(&pdisp);
   exit(hr);

   hr = pdisp->QueryInterface(__uuidof(ICatalogObject), (void**)&pobj);
   exit(hr);

   VariantInit(&var);

   var.vt = VT_BSTR;
   var.bstrVal = SysAllocString(m_pwzAppName);
   if (!var.bstrVal)
      exit(hr=E_OUTOFMEMORY);

   hr = pobj->put_Value(g_bNAME, var);
   exit(hr);

   bname = L"Activation";
   VariantClear(&var);
   var.vt = VT_BOOL;
   var.boolVal = (short) bServer;

   hr = pobj->put_Value(bname,var);
   exit(hr);

   bname = L"QueuingEnabled";
   var.vt = VT_BOOL;
   var.boolVal = (short) bQueued;
   
   hr = pobj->put_Value(bname,var);
   exit(hr);

   bname = L"QueueListenerEnabled";
   var.vt = VT_BOOL;
   var.boolVal = (short) bQueued;

   hr = pobj->put_Value(bname,var);
   exit(hr);      

   hr = pcoll->SaveChanges(&lret);

exit:
   VariantClear(&var);
   SAFERELEASE(pcoll);
   SAFERELEASE(pobj);
   SAFERELEASE(pdisp);

   ASSERT_AND_RETURN(hr);
   return hr;
}

//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::DeleteApp
// Purppose    : Deletes a COM+ Application and its components
// Parameters  :
// Return Value : HRESULT
///////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::DeleteApp()
{
   HRESULT           hr = E_FAIL;
   ICatalogCollection   *pApps = NULL;
   ICatalogObject    *pApp = NULL;
   _bstr_t           bappname;

   if (!m_pwzAppName || !m_pAdminCatalog)
      return E_POINTER;

   bappname = m_pwzAppName;

   hr = GetCollection(m_pAdminCatalog, g_bAPP, &pApps);
   exit(hr);
   hr = RemoveNamedObjectFromCollection(pApps, bappname, APPLICATION); 
   m_pwzAppName = NULL;
exit:
   SAFERELEASE(pApps);
   SAFERELEASE(pApp);
   ASSERT_AND_RETURN(hr);

   return hr;
}

//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::InstallComponent
// Purppose    : Installs a component
// Parameters  :
//             : WCHAR* pwzcompdllname    - Path of the dll which exports this 
//                                          component
// Return Value : HRESULT
///////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::InstallComponent(WCHAR* pwzcompdllname)
{
   _bstr_t        bappname, bdllname;

   if (!m_pwzAppName || !pwzcompdllname || !m_pAdminCatalog)
      return E_INVALIDARG;

   bappname = m_pwzAppName;
   bdllname = pwzcompdllname;
   
   HRESULT hres = m_pAdminCatalog->InstallComponent(bappname, pwzcompdllname, 0, 0);
      
   ASSERT_AND_RETURN(hres);

   return hres;
}

//////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::RemoveComponent
// Purppose    : Removes the specified component from the application
// Parameters  :
//             : WCHAR* progid - Name of the component
// Return Value : HRESULT
///////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::RemoveComponent(WCHAR* progid)
{
   ICatalogCollection      *pApps=0, *pComps=0;
   ICatalogObject       *pApp=0;
   HRESULT              hr =0;

   if (!m_pwzAppName || !m_pAdminCatalog || !progid)
      return E_INVALIDARG;


   hr = GetObjectAndCollection(m_pAdminCatalog, g_bAPP, _bstr_t(m_pwzAppName), APPLICATION, &pApps, &pApp);
   exit(hr);

   hr = GetCollection(pApps, pApp, g_bCOMP, &pComps);
   exit(hr);

   hr = RemoveNamedObjectFromCollection(pComps, _bstr_t(progid), COMPONENT);

exit:
   SAFERELEASE(pApps);
   SAFERELEASE(pComps);
   SAFERELEASE(pApp);
   
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::AddEventClass
// Purppose    : Installs an EventClass component under the application
// Parameters  :
//             : WCHAR* pwzdllname      - Dll path that has DllRegisterServer exported
//             : WCHAR* pwztlbpath      - TLB path that contains Firing Interface definitions
//             : WCHAR* pwzCompName     - Name of the component
//             : WCHAR* pwzPublisherID  - Name of the Publisher
//             : WCHAR* pwzFilterCLSID  - CLSID (in string from) of the filter component
//             : BOOL   fAllowinproc    - Flag that sets the "AllowInProcSubscribers" property
//             : BOOL   fFireinparallel - Flag that decides if it is serial or parallel firing
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::AddEventClass(WCHAR* pwzdllname, WCHAR* pwztlbpath, WCHAR* pwzCompName, WCHAR* pwzPublisherID, WCHAR* pwzFilterCLSID, BOOL fAllowinproc, BOOL fFireinparallel)
{
   HRESULT              hr = 0;
   ICatalogCollection      *pApps=0, *pComps=0;
   ICatalogObject       *pApp=0, *pComp=0;
   long              lret =0;
   _bstr_t              bapp, bdll("\0"), btlb("\0");
   BOOL              fCreate=FALSE, fComplete = FALSE;

   if (!pwzdllname || !m_pAdminCatalog || !m_pwzAppName || !pwzCompName)
      return E_INVALIDARG;

   if (pwzdllname)
      bdll = pwzdllname;

   if (pwztlbpath)
      btlb = pwztlbpath;
   
   bapp = m_pwzAppName;

   hr = m_pAdminCatalog->InstallEventClass(bapp, bdll, btlb,0);
   exit(hr);

   fCreate = TRUE;

   hr = GetObjectAndCollection(m_pAdminCatalog, g_bAPP, bapp, APPLICATION, &pApps, &pApp);
   exit(hr);

   hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, _bstr_t(pwzCompName), COMPONENT, &pComps, &pComp);
   exit(hr);

   hr = SetStringProperty(pComp, g_bPUBID, pwzPublisherID);
   exit(hr);

   hr = SetStringProperty(pComp, g_bFILTER, pwzFilterCLSID);
   exit(hr);

   hr = SetBoolProperty(pComp, g_bALLOWINPROC, fAllowinproc);
   exit(hr);

   hr = SetBoolProperty(pComp, g_bFIREINPARALLEL, fFireinparallel);
   exit(hr);

   hr = pComps->SaveChanges(&lret);
   exit(hr);

   fComplete = TRUE;
   
exit:
   if (fCreate && !fComplete)
      RemoveComponent(pwzCompName);

   SAFERELEASE(pComp);
   SAFERELEASE(pApps);
   SAFERELEASE(pComps);
   SAFERELEASE(pApps);

   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::AddSubscription
// Purppose    : Creates a Persistent Subscription
// Parameters  :
//             : WCHAR* pwzProgID   - Name of the component for which the subscription to be created
//             : WCHAR* pwzSubName  - Name of the subscription
//             : WCHAR* pwzECID     - EventClass ID
//             : WCHAR* pwzPubID    - Name of the Publisher
//             : WCHAR* pwzIID      - Interface Name (in string form)
//             : WCHAR* pwzMethod   - Method in the interface to subscribe
//             : WCHAR* pwzCriteria - Filter Criteria (used in parameterized filtering)
//             : WCHAR* pwzPubProp  - Publisher Property Name
//             : WCHAR* pwzPubVal   - Publisher Property Value
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::AddSubscription(WCHAR* pwzProgID, WCHAR* pwzSubName, WCHAR* pwzECID, WCHAR* pwzPubID, WCHAR* pwzIID, WCHAR* pwzMethod, WCHAR* pwzCriteria, WCHAR* pwzPubProp, WCHAR* pwzPubVal)
{
   HRESULT           hr=0;
   ICatalogObject    *pApp=0,  *pComp=0, *pSub=0, *pProp=0;
   ICatalogCollection  *pApps=0, *pComps=0, *pSubs=0, *pProps=0;
   IDispatch         *pDisp=0;
   _bstr_t           bcomp, bapp;
   BOOL           fCreate=FALSE, fComplete = FALSE;
   long           lret;

   if (!m_pAdminCatalog || !pwzProgID || !m_pwzAppName)
      return E_INVALIDARG;

   bcomp = pwzProgID;
   bapp  = m_pwzAppName;

   hr = GetObjectAndCollection(m_pAdminCatalog, g_bAPP, bapp, APPLICATION, &pApps, &pApp);
   exit(hr);

   hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, bcomp, COMPONENT, &pComps, &pComp);
   exit(hr);

   hr = GetCollection(pComps, pComp, g_bSUB, &pSubs);
   exit(hr);

   hr = pSubs->Add(&pDisp);
   exit(hr);

   hr = pDisp->QueryInterface(__uuidof(IDispatch), (LPVOID*)&pSub);
   exit(hr);

   hr = SetStringProperty(pSub, g_bNAME, pwzSubName);
   exit(hr);

   hr = SetStringProperty(pSub, g_bECID, pwzECID);
   exit(hr);

   hr = SetStringProperty(pSub, g_bPUBID, pwzPubID);
   exit(hr);

   hr = SetStringProperty(pSub, g_bIID, pwzIID);
   exit(hr);

   hr = SetStringProperty(pSub, g_bMETHOD, pwzMethod);
   exit(hr);

   hr = SetStringProperty(pSub, g_bCRITERIA, pwzCriteria);
   exit(hr);

   hr = pSubs->SaveChanges(&lret);
   exit(hr);

   fCreate = TRUE;

   if (pwzPubProp && pwzPubVal)
   {
      hr = GetCollection(pSubs, pSub, g_bPUBPROP, &pProps);
      exit(hr);

      SAFERELEASE(pDisp);
      hr = pProps->Add(&pDisp);
      exit(hr);

      hr = pDisp->QueryInterface(__uuidof(ICatalogObject), (LPVOID*)&pProp);
      exit(hr);

      hr = SetStringProperty(pProp, g_bNAME, pwzPubProp);
      exit(hr);

      hr = SetStringProperty(pProp, g_bVALUE, pwzPubVal);
      exit(hr);

      hr = pProps->SaveChanges(&lret);
      exit(hr);
   }

   fComplete = TRUE; 

exit:
   if (fCreate && !fComplete)
      RemoveSubscription(pwzProgID, pwzSubName);

   SAFERELEASE(pDisp);
   SAFERELEASE(pProp);
   SAFERELEASE(pProps);
   SAFERELEASE(pDisp);
   SAFERELEASE(pSub);
   SAFERELEASE(pSubs);
   SAFERELEASE(pComp);
   SAFERELEASE(pComps);
   SAFERELEASE(pApp);
   SAFERELEASE(pApps);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::RemoveSubscription
// Purppose    : Removes a Persistent Subscription
// Parameters  :
//             : WCHAR* pwzProgID  - Name of the component for which the subscription to be deleted
//             : WCHAR* pwzSubName - Name of the subscription
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::RemoveSubscription(WCHAR* pwzProgID, WCHAR* pwzSubName)
{
   HRESULT           hr =0;
   ICatalogObject    *pApp =0, *pComp=0;
   ICatalogCollection   *pApps=0, *pComps=0, *pSubs=0;

   if (!m_pAdminCatalog || !m_pwzAppName || !pwzProgID || !pwzSubName)
      return E_INVALIDARG;

   hr = GetObjectAndCollection(m_pAdminCatalog, g_bAPP, _bstr_t(m_pwzAppName), APPLICATION, &pApps, &pApp);
   exit(hr);

   hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, _bstr_t(pwzProgID), COMPONENT, &pComps, &pComp);
   exit(hr);

   hr = GetCollection(pComps, pComp, g_bSUB, &pSubs);
   exit(hr);

   hr = RemoveNamedObjectFromCollection(pSubs, _bstr_t(pwzSubName), SUBSCRIPTION);

exit:
   SAFERELEASE(pSubs);
   SAFERELEASE(pApp);
   SAFERELEASE(pApps);

   return hr;
}

HRESULT CApp::AddTransientSubscription(WCHAR* pwzSubName,CLSID clsidEventClass,IID iidInterface, IUnknown *punk)
{
   return AddTransientSubscription(pwzSubName,clsidEventClass,NULL,iidInterface,punk,NULL,NULL,NULL,NULL);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::AddTransientSubscription
// Purppose    : Adds a Persistent Subscription
// Parameters  :
//             : WCHAR* pwzSubName  - Name of the subscription
//             : CLSID  clsidEventClass - EventClass ID
//             : WCHAR* pwzPubID    - Name of the Publisher
//             : IID    iidInterface - Interface ID
//             : IUnknown* punk     - IUnknown* of the subscriber
//             : WCHAR* pwzMethod   - Method in the interface to subscribe
//             : WCHAR* pwzCriteria - Filter Criteria (used in parameterized filtering)
//             : WCHAR* pwzPubProp  - Publisher Property Name
//             : WCHAR* pwzPubVal   - Publisher Property Value
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::AddTransientSubscription(WCHAR* pwzSubName,CLSID clsidEventClass, WCHAR* pwzPubID,IID iidInterface, IUnknown *punk, WCHAR* pwzMethod, WCHAR* pwzCriteria, WCHAR* pwzPubProp, WCHAR* pwzPubVal)
{
   HRESULT            hres = S_OK;
   ICatalogObject     *pSub=0, *pProp=0;
   ICatalogCollection *pSubs=0, *pProps=0;
   IDispatch          *pDisp=0;
   long               lret=0;
   BOOL               fCreate = FALSE, fComplete = FALSE;
   WCHAR*  pwsEventClass = NULL;
   WCHAR*  pwsInterfaceIID = NULL;

   if (!pwzSubName || !punk)
      return E_INVALIDARG;

   hres = ::StringFromCLSID(clsidEventClass, &pwsEventClass);
   ASSERT_AND_RETURN(hres);

   if(IID_NULL == iidInterface)
   {
      pwsInterfaceIID = NULL;
   }
   else
   {
      hres = ::StringFromCLSID(iidInterface, &pwsInterfaceIID);
      ASSERT_AND_RETURN(hres);
   }

   ICOMAdminCatalog* pAdminCat = NULL;

   hres = ::CoCreateInstance(__uuidof(COMAdminCatalog), NULL, CLSCTX_SERVER, __uuidof(ICOMAdminCatalog),(LPVOID*) &pAdminCat);
   ASSERT_AND_RETURN(hres);

   hres = GetCollection(pAdminCat, g_bTRANSSUB, &pSubs);
   exit(hres);
   pAdminCat->Release();

   hres = pSubs->Add(&pDisp);
   exit(hres);

   hres = pDisp->QueryInterface(__uuidof(ICatalogObject), (LPVOID*)&pSub);
   exit(hres);

   hres = SetStringProperty(pSub, g_bNAME, pwzSubName);
   exit(hres);

   hres = SetStringProperty(pSub, g_bECID, pwsEventClass);
   exit(hres);

   hres = SetStringProperty(pSub, g_bPUBID, pwzPubID);
   exit(hres);

   hres = SetStringProperty(pSub, g_bIID, pwsInterfaceIID);
   exit(hres);

   hres = SetIUnknownProperty(pSub, g_bSUBINT, punk);
   exit(hres);

   hres = SetStringProperty(pSub, g_bMETHOD, pwzMethod);
   exit(hres);

   hres = SetStringProperty(pSub, g_bCRITERIA, pwzCriteria);
   exit(hres);

   hres = pSubs->SaveChanges(&lret);
   exit(hres);

   fCreate = TRUE;

   if (pwzPubProp && pwzPubVal)
   {

      hres = GetCollection(pSubs, pSub, g_bTRANSPUBPROP, &pProps);
      exit(hres);

      SAFERELEASE(pDisp);
      hres = pProps->Add(&pDisp);
      exit(hres);

      hres = pDisp->QueryInterface(__uuidof(ICatalogObject), (LPVOID*)&pProp);
      exit(hres);

      hres = SetStringProperty(pProp, g_bNAME, pwzPubProp);
      exit(hres);
      
      hres = SetStringProperty(pProp, g_bVALUE, pwzPubVal);
      exit(hres);

      hres = pProps->SaveChanges(&lret);
      exit(hres);
   }

   fComplete = TRUE;
   
exit:
   if (fCreate && !fComplete)
      RemoveTransientSubscription(pwzSubName);
   SAFERELEASE(pDisp);
   SAFERELEASE(pProp);
   SAFERELEASE(pProps);
   SAFERELEASE(pSub);
   SAFERELEASE(pSubs);
      
   ASSERT_AND_RETURN(hres);

   return hres;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::RemoveTransientSubscription
// Purppose    : Remolves a Persistent Subscription
// Parameters  :
//             : WCHAR* pwzSubName - Name of the subscription
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::RemoveTransientSubscription(WCHAR* pwzName)
{
   HRESULT              hr =0;
   ICatalogCollection      *pSubs=0;

   if (!pwzName)
      return E_INVALIDARG;

   ICOMAdminCatalog* pAdminCat = NULL;

   hr = ::CoCreateInstance(__uuidof(COMAdminCatalog), NULL, CLSCTX_SERVER, __uuidof(ICOMAdminCatalog),(LPVOID*) &pAdminCat);
   ASSERT(SUCCEEDED(hr));

   hr = GetCollection(pAdminCat, g_bTRANSSUB, &pSubs);
   exit(hr);

   pAdminCat->Release();

   hr = RemoveNamedObjectFromCollection(pSubs, _bstr_t(pwzName), TRANSIENTSUB);
exit:
   SAFERELEASE(pSubs);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : CApp::SetProperty
// Purppose    : Sets a property on any component (Application, EventClass, Component, Persistent 
//             : subscription, Transient subscription, Persistent sub. property bag, Trans sub. 
//             : property bag)
// Parameters  :
//             : ADMINTYPE type      - Type of the component
//             : WCHAR* pwztypename  - Value depends on the type
//             : WCHAR* pwztypename2 - Value depends on the type
//             : WCHAR* pwzPropName  - Name of the Property
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT CApp::SetProperty(ADMINTYPE type, WCHAR* pwztypename, WCHAR* pwztypename2, WCHAR* pwzPropName, VARIANT& var)
{
   HRESULT             hr=0;
   long                lindex;
   _bstr_t             bname;
   VARIANT             varname;
   ICatalogObject     *pApp=0, *pComp=0, *pSub=0, *pProp=0, *pFinal=0;
   ICatalogCollection *pApps=0, *pComps=0, *pSubs=0, *pProps=0, *pFinals=0;
   IDispatch          *pDisp=0;
   long                lret =0;
   

   // ADMINTYPE               pwztypename         pwztypename2         
   //========================================================================
   // APPLICATION             Ignored             Ignored              
   // COMPONENT               progid of the comp  Ignored           
   // SUBSCRIPTION            progid of the comp  name of the sub      
   // TRANSIENTSUB            name of the sub     Ignored           
   // PUBPROP                 progid of the comp  name of the sub      
   // TRANSIENTPUBPROP        name of the sub     Ignored
   if (!m_pAdminCatalog || !m_pwzAppName || var.vt == VT_EMPTY)
      return E_INVALIDARG;

   hr = GetObjectAndCollection(m_pAdminCatalog, g_bAPP, _bstr_t(m_pwzAppName), APPLICATION, &pApps, &pApp);
   exit(hr);
   bname = pwzPropName;

   switch(type)
   {
   case APPLICATION:
      if (!pwzPropName)
         exit(hr= E_INVALIDARG);
      pFinal = pApp;
      pFinals = pApps;
      break;

   case COMPONENT:
      if (!pwztypename || !pwzPropName)
         exit(hr= E_INVALIDARG);

      hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, _bstr_t(pwztypename), COMPONENT, &pComps, &pComp);
      exit(hr);
      pFinal = pComp;
      pFinals = pComps;
      break; 

   case SUBSCRIPTION:
      if (!pwztypename || !pwztypename2 || !pwztypename)
         exit(hr= E_INVALIDARG);

      hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, _bstr_t(pwztypename), COMPONENT, &pComps, &pComp);
      exit(hr);

      hr = GetObjectAndCollection(pComps, pComp, g_bSUB, _bstr_t(pwztypename2), SUBSCRIPTION, &pSubs, &pSub);
      exit(hr);

      pFinal = pSub;
      pFinals = pSubs;
      break;

   case TRANSIENTSUB:
      if (!pwzPropName)
         exit(hr= E_INVALIDARG);
      hr = GetObjectAndCollection(m_pAdminCatalog, g_bTRANSSUB, _bstr_t(pwztypename), TRANSIENTSUB, &pSubs, &pSub);
      exit(hr);

      pFinal = pSub;
      pFinals = pSubs;
      break;

   case PUBPROP:
      if (!pwzPropName || !pwztypename || !pwztypename2)
         exit(hr= E_INVALIDARG);

      hr = GetObjectAndCollection(pApps, pApp, g_bCOMP, _bstr_t(pwztypename), COMPONENT, &pComps, &pComp);
      exit(hr);

      hr = GetObjectAndCollection(pComps, pComp, g_bSUB, _bstr_t(pwztypename2), SUBSCRIPTION, &pSubs, &pSub);
      exit(hr);

      hr = GetCollection(pSubs, pSub, g_bPUBPROP, &pProps);
      exit(hr);

      hr = GetNamedObjectFromCollection(pProps, g_bNAME, &pProp, &lindex, PUBPROP, _bstr_t(pwzPropName));
      if SUCCEEDED(hr)  // Change the VALUE
         bname = g_bVALUE;
      else if  (FAILED(hr) && (lindex == -1))         // Create a new NAME/VALUE property
      {
         hr = pProps->Add(&pDisp);
         exit(hr);

         hr = pDisp->QueryInterface(__uuidof(IDispatch), (LPVOID*)&pProp);
         exit(hr);

         varname.bstrVal = SysAllocString(pwzPropName);
         if (!varname.bstrVal)
            exit(hr=E_OUTOFMEMORY);
         varname.vt = VT_BSTR;

         hr = pProp->put_Value(g_bNAME, varname);
         exit(hr);

         bname = g_bVALUE;
         pFinal = pProp;
         pFinals = pProps;
      }
      else
         exit(hr);
      pFinal = pProp;
      pFinals = pProps;
      break;

   case TRANSIENTPUBPROP:
      if (!pwzPropName || !pwztypename)
         exit(hr= E_INVALIDARG);

      hr = GetObjectAndCollection(m_pAdminCatalog, g_bTRANSSUB, _bstr_t(pwztypename), TRANSIENTSUB, &pSubs, &pSub);
      exit(hr);

      hr = GetCollection(pSubs, pSub, g_bTRANSPUBPROP, &pProps);
      exit(hr);

      hr = GetNamedObjectFromCollection(pProps, g_bNAME, &pProp, &lindex, TRANSIENTPUBPROP, _bstr_t(pwzPropName));
      if SUCCEEDED(hr)  // Change the VALUE
         bname = g_bVALUE;
      else if  (FAILED(hr) && (lindex == -1))      // Create a new NAME/VALUE property
      {
         hr = pProps->Add(&pDisp);
         exit(hr);

         hr = pDisp->QueryInterface(__uuidof(IDispatch), (LPVOID*)&pProp);
         exit(hr);

         varname.bstrVal = SysAllocString(pwzPropName);
         if (!varname.bstrVal)
            exit(hr=E_OUTOFMEMORY);
         varname.vt = VT_BSTR;

         hr = pProp->put_Value(g_bNAME, varname);
         exit(hr);

         bname = g_bVALUE;
         pFinal = pProp;
         pFinals = pProps;
      }
      else
         exit(hr);
      pFinal = pProp;
      pFinals = pProps;
      break;

   default:
      exit(hr= E_INVALIDARG);
   }

   hr = pFinal->put_Value(bname, var);
   exit(hr);

   hr = pFinals->SaveChanges(&lret);

exit:
   SAFERELEASE(pDisp);
   SAFERELEASE(pProp);
   SAFERELEASE(pProps);
   SAFERELEASE(pSubs);
   SAFERELEASE(pSub);
   SAFERELEASE(pApp);
   SAFERELEASE(pApps);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetCollection
// Purppose    : Gets a collection from the Catalogue
// Parameters  :
//             : ICOMAdminCatalog*    pCat      - Admin Catalog object
//             : BSTR                 szCollName - Collection Name
//             : ICatalogCollection** ppCol     - Collection returned
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT GetCollection(ICOMAdminCatalog* pCat, BSTR szCollName, ICatalogCollection** ppCol)
{
   HRESULT hr = E_INVALIDARG;
   CComBSTR bstr;
   IDispatch* pDisp = NULL;
   if (!pCat || !szCollName || !ppCol) 
      goto exit;
   bstr = szCollName;
   hr = pCat->GetCollection(bstr.m_str, &pDisp);
   exit(hr);
   hr = pDisp->QueryInterface(__uuidof(ICatalogCollection), (LPVOID*)ppCol);
   exit(hr);
   hr = (*ppCol)->Populate();
exit:
   SAFERELEASE(pDisp);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetCollection
// Purppose    : Gets a collection from the Catalogue
// Parameters  :
//             : ICatalogCollection*  pParentColl - Parent Collection
//             : ICatalogObject*      pCatObj     - Parent Object
//             : BSTR                 szCollName   - Collection Name
//             : ICatalogCollection** ppCol       - Collection returned
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT GetCollection(ICatalogCollection* pParentColl, ICatalogObject* pCatObj, BSTR szCollName, ICatalogCollection** ppCol)
{
   HRESULT hr = E_INVALIDARG;
   CComBSTR bstr;
   VARIANT var;
   IDispatch* pDisp = NULL;
   VariantInit(&var);
   if (!pParentColl || !pCatObj || !szCollName || !ppCol)
      goto exit;
   hr = pCatObj->get_Key(&var);
   exit(hr);
   bstr = szCollName;
   hr = pParentColl->GetCollection(bstr.m_str, var, &pDisp);
   exit(hr);
   hr = pDisp->QueryInterface(__uuidof(ICatalogCollection), (LPVOID*)ppCol);
   exit(hr);
   hr = (*ppCol)->Populate();
exit:
   VariantClear(&var);
   if(pDisp)
      pDisp->Release();
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetObjectAndCollection
// Purppose    : Gets a collection from the Catalogue
// Parameters  :
//             : ICOMAdminCatalog*    pCat        - Admin Catalog object
//             : BSTR                 szCollName   - Collection Name
//             : BSTR                 szObjectName - Object name
//             : ADMINTYPE            type         - Type of the object
//             : ICatalogCollection** ppCol       - Collection returned
//             : ICatalogObject*      ppObj       - Object returned
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT GetObjectAndCollection(ICOMAdminCatalog* pCat, BSTR szCollName, BSTR szObjectName, ADMINTYPE type, ICatalogCollection** ppCol, ICatalogObject** ppObj)
{
   HRESULT hr = E_FAIL;
   hr = GetCollection(pCat, szCollName, ppCol);
   exit(hr);
   hr = GetNamedObjectFromCollection(*ppCol, szObjectName, ppObj, NULL, type);
exit:
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetObjectAndCollection
// Purppose    : Gets a collection from the Catalogue
// Parameters  :
//             : ICatalogCollection*  pParentColl   - Parent Collection
//             : ICatalogObject*      pParentObject - Parent Object
//             : BSTR                 szCollName     - Collection Name
//             : BSTR                 szObjectName   - Object name
//             : ADMINTYPE            type           - Type of the object
//             : ICatalogCollection** ppCol         - Collection returned
//             : ICatalogObject*      ppObj         - Object returned
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT GetObjectAndCollection(ICatalogCollection* pParentColl, ICatalogObject* pParentObject, BSTR szCollName, BSTR szObjectName, ADMINTYPE type, ICatalogCollection** ppCol, ICatalogObject** ppObj)
{
   HRESULT hr = E_FAIL;
   hr = GetCollection(pParentColl, pParentObject, szCollName, ppCol);   
   exit(hr);
   hr = GetNamedObjectFromCollection(*ppCol, szObjectName, ppObj, NULL, type);
exit:
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : RemoveNamedObjectFromCollection
// Purppose    : Removes an object from a collection
// Parameters  :
//             : ICatalogCollection* pParentColl - Collection from the object has to be removed
//             : BSTR                szObjectName - Object name
//             : ADMINTYPE           type         - type of the object
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT RemoveNamedObjectFromCollection(ICatalogCollection* pCol, BSTR szObjectName, ADMINTYPE type)
{
   HRESULT hr = E_FAIL;
   ICatalogObject* pCatObj = NULL;
   LONG i, lTotal;

   if (wcscmp(szObjectName, SysAllocString(L"*"))) 
   {
      hr = GetNamedObjectFromCollection(pCol, szObjectName, &pCatObj, &i, type);
      exit(hr);
      hr = pCol->Remove(i);
      exit(hr);
      hr = pCol->SaveChanges(&i);
   }
   else
   {
      hr = pCol->get_Count(&lTotal);
      exit(hr);
      for (i=lTotal-1; i >= 0; i--)
      {
         hr = pCol->Remove(i);
         exit(hr);
         hr = pCol->SaveChanges(&i);
         exit(hr);
      }
   }

exit:
   if(pCatObj)
      pCatObj->Release();
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetNamedObjectFromCollection
// Purppose    : Gets an object from a collection
// Parameters  :
//             : ICatalogCollection* pParentColl - Collection from the object has to be retrieved
//             : BSTR                bstrObjName  - Object name
//             : ADMINTYPE           type         - type of the object
//             : ICatalogObject**    ppObj        - return Object
//             : LONG*               plIndex      - Index of the object (-1, if not found)
//             : ADMINTYPE           type         - type of the object
//             : BSTR                bstrPropname - name of the property
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT GetNamedObjectFromCollection(ICatalogCollection* pCol, BSTR bstrObjName, ICatalogObject** ppObj, LONG* plIndex, ADMINTYPE type, BSTR bstrPropname)
{
   HRESULT hr = E_FAIL;
   LONG lCount = 0;
   LONG i = 0;
   ICatalogObject* pObj = NULL;
   BSTR bstrName;
   VARIANT varProp;
   BOOL  fFound = FALSE;

   VariantInit(&varProp);
   if (!pCol || !bstrObjName || !ppObj)   
      return E_INVALIDARG;
   
   if (!bstrPropname)
   {
      hr = GetObjectNameProperty(type, &bstrName);
      exit(hr);
   }
   else
      bstrName = bstrPropname;

   *ppObj = NULL;

   hr = pCol->Populate();
   exit(hr);

   hr = pCol->get_Count(&lCount);
   exit(hr);

   if (lCount == 0)
   {
      *plIndex = -1; // To indicate that the object is not found in the collection
      exit(hr=E_FAIL);
   }
   
   // Loop through 
   for (i = 0; i < lCount; i++)
   {
      VariantClear(&varProp);

      // Get the next item
      hr = pCol->get_Item(i, (IDispatch**)&pObj);
      exit(hr);
      
      // Retrieve it's name property
      hr = pObj->get_Value((BSTR)bstrName, &varProp);
      exit(hr);

      // Check it's name property for a match
      if (wcscmp(bstrObjName, varProp.bstrVal) == 0)
      {
         // Found it!
         *ppObj = pObj;
         pObj->AddRef();
         VariantClear(&varProp);
         // does user want index of the just-found object?
         if (plIndex)
            *plIndex = i;
         fFound = TRUE;
         hr = S_OK;
         goto exit;
      }
       SAFERELEASE(pObj);
   }
   if (plIndex)
      *plIndex = -1; // To indicate that the object is not found in the collection
exit:
   SysFreeString(bstrName);
   // If we got here then we didn't find it
   SAFERELEASE(pObj);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : GetObjectNameProperty
// Purppose    : Gets the property name of the object
// Parameters  :
//             : ADMINTYPE type  - type of the object
//             : BSTR*     pbstr - name
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT GetObjectNameProperty(ADMINTYPE type, BSTR* pbstr)
{  
   if (!pbstr)
      return E_INVALIDARG;

   switch (type)
   {  
   case APPLICATION:
   case SUBSCRIPTION:
   case TRANSIENTSUB:
      *pbstr = SysAllocString(g_wczNAME);
      break;
   case COMPONENT:
      *pbstr = SysAllocString(g_wczProgIDProp);
      break;
   default:
      assert(FALSE);
      break;
   }
   return S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : SetStringProperty
// Purppose    : Sets the named string property of the object
// Parameters  :
//             : ICatalogObject* pcat  -  Pointer to the object
//             : BSTR            bPROP -  Name of the property
//             : WCHAR*          pval  -  Value of the property
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT SetStringProperty(ICatalogObject *pcat, BSTR bPROP, WCHAR* pval)
{
   HRESULT hr;
   VARIANT var;
   if (!pcat)
      return E_INVALIDARG;

   if (!pval)
      return S_FALSE;

   VariantClear(&var);

   var.bstrVal = SysAllocString(pval);
   if (!var.bstrVal)
      exit(hr=E_OUTOFMEMORY);
   var.vt = VT_BSTR;

   hr = pcat->put_Value(bPROP, var);
   exit(hr);

exit:
   VariantClear(&var);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : SetIUnknownProperty
// Purppose    : Sets the named IUnknown property of the object
// Parameters  :
//             : ICatalogObject *pcat  -  Pointer to the object
//             : BSTR            bPROP -  Name of the property
//             : IUnknown*       punk  -  Value of the property
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT SetIUnknownProperty(ICatalogObject *pcat, BSTR bPROP, IUnknown* punk)
{
   HRESULT hr;
   VARIANT var;
   if (!punk || !pcat)
      return E_INVALIDARG;

   VariantClear(&var);

   var.punkVal = punk;
   var.vt = VT_UNKNOWN;
   punk->AddRef();   // We'll be caling VariantClear

   hr = pcat->put_Value(bPROP, var);
   exit(hr);

exit:
   VariantClear(&var);
   return hr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Name        : SetBOOLProperty
// Purppose    : Sets the named BOOL property of the object
// Parameters  :
//             : ICatalogObject *pcat -  Pointer to the object
//             : BSTR bPROP           -  Name of the property
//             : BOOL fFlag           -  Value of the property
// Return Value : HRESULT
////////////////////////////////////////////////////////////////////////////////////////////////////  
HRESULT SetBoolProperty(ICatalogObject *pcat, BSTR bPROP, BOOL fFlag)
{
   HRESULT  hr;
   VARIANT var;

   if (!pcat)
      return E_INVALIDARG;

   VariantClear(&var);

   var.boolVal = (short) fFlag;
   var.vt = VT_BOOL;

   hr = pcat->put_Value(bPROP, var);
   exit(hr);

exit:
   VariantClear(&var);
   return hr;
}
